import re

import json


class CodeContentClassification:
    Label_Unknown = -1
    Label_Code = 0
    Label_Stacktrace = 1
    Label_XML_or_JSON = 2
    Label_Number = 3
    Label_Text = 4
    Label_Command = 5
    label_code_2_name = {
        Label_Unknown: "Unknown",
        Label_Code: "Code",
        Label_Stacktrace: "Stacktrace",
        Label_XML_or_JSON: "XML_or_JSON",
        Label_Number: "Number",
        Label_Text: "Text",
        Label_Command: "Command",
    }

    def __init__(self):
        # 注释
        self.comment_pattern = re.compile(r'/\*\*.+?\*/', re.DOTALL)
        # 在文本后面直接加括号，没有空格将其分开
        self.brackets_pattern = re.compile(r'\S\(', re.DOTALL)
        # 两个单词之间的点
        self.point_pattern = re.compile(r'\S\.\S', re.DOTALL)
        # 两个单词之间的箭头
        self.arrow_pattern = re.compile(r'\S\.\S', re.DOTALL)
        self.CAMEL_CASE_TEST_RE = re.compile(r'^[a-zA-Z]*([a-z]+[A-Z]+|[A-Z]+[a-z]+)[a-zA-Z\d]*$')
        # 日期
        self.data_pattern = re.compile(r'\d{2}-\d{1,2} \d{1,2}:\d{1,2}:\d{1,2}')
        # 错误行:247
        self.error_row_pattern = re.compile(r':\d+\)')
        #     [INFO]
        self.info_pattern = re.compile(r'\[INFO\] ')
        #     [ERROR]
        self.error_pattern = re.compile(r'\[ERROR\] ')
        #     [WARNING]
        self.warning_pattern = re.compile(r'\[WARNING\] ')
        self.api_pattern = re.compile(r'\w+\.\w+\.\w+')
        self.xml_end = re.compile(r'</\w+>')

    def so_text_label(self, text):
        pass

    def count_upper(self, word):
        c = 0
        for e in word:
            if e.isupper():
                c += 1
        return c

    def all_camel_case(self, text):
        """
        Checks if a string is formatted as camel case.
        A string is considered camel case when:
        - it's composed only by letters ([a-zA-Z]) and optionally numbers ([0-9])
        - it contains both lowercase and uppercase letters
        - it does not start with a number
        :param text: String to test.
        :type text: str
        """
        have = set()
        af = re.split('[.( \s]', text)
        for each in af:
            if len(each) > 0 and each[0].isupper() and self.count_upper(each) == 1:
                continue
            if bool(self.CAMEL_CASE_TEST_RE.search(each)):
                have.add(each)
        return have

    def check_is_trace(self, text: str):
        # 含有这种日期格式 06-03 15:05:29
        data_len = len(self.data_pattern.findall(text))
        af = text.splitlines()
        line_num = 1.0 * len(af)
        if data_len / line_num >= 0.4:
            return True
        error_row_len = len(self.error_row_pattern.findall(text))
        if error_row_len / line_num >= 0.3:
            return True
        log_len = len(self.info_pattern.findall(text))
        log_len += len(self.error_pattern.findall(text))
        log_len += len(self.warning_pattern.findall(text))
        if log_len / line_num >= 0.3:
            return True
        for each in af:
            if text.find("Error:") >= 0 and len(self.api_pattern.findall(each)) >= 1:
                return True
            if text.find("Exception:") >= 0 and len(self.api_pattern.findall(each)) >= 1:
                return True
        return False

    def check_json(self, text: str):
        if self.is_json(text):
            if not self.check_nums(text):
                return True

        return False

    def is_json(self, myjson):
        try:
            json.loads(myjson)
        except ValueError:
            return False
        return True

    def check_nums(self, text: str):
        text = "".join(text.split())
        all_size = 1.0 * len(text)
        count = 0
        for c in text:
            if c.isdigit():
                count += 1
        if count / all_size >= 0.32:
            return True
        return False

    def check_xml(self, text: str):
        if text.find("</groupId>") >= 0:
            return True
        if text.find("dependencies {") >= 0:
            return True
        if text.find("-XX:") >= 0 and text.find("<") >= 0 and text.find(">") >= 0:
            return True

        text = text.replace("<p>", "").replace("</p>", "")
        depth = self.maxDepth(text, '<', '>')

        af = text.splitlines()
        end_count = len(self.xml_end.findall(text))

        if 0 < depth <= end_count:
            return True

        return False

    def check_is_command(self, text: str):
        if text.count("#") > 3:
            return False
        text = text.strip()
        if text.find("java -") >= 0:
            return True
        if text.find("apt-") >= 0:
            return True
        if text.find("sudo ") >= 0:
            return True
        if text.find("/usr/bin/") >= 0 or text.find("/var/lib/") >= 0:
            return True

        return False

    def get_label(self, text: str):

        if self.check_is_trace(text):
            return CodeContentClassification.Label_Stacktrace
        if self.check_xml(text) or self.check_json(text):
            return CodeContentClassification.Label_XML_or_JSON
        if self.check_is_command(text):
            return CodeContentClassification.Label_Command
        if self.check_is_code(text):
            return CodeContentClassification.Label_Code
        if self.check_nums(text):
            return CodeContentClassification.Label_Number
        return CodeContentClassification.Label_Text

    def check_is_code(self, text: str):
        # 行尾的分号。
        # 在文本后面直接加括号，没有空格将其分开： myFunc()
        # 两个单词之间的点或箭头： foo.bar = ptr->val
        # 花括号，方括号的存在： while (true) { bar[i]; }
        # 存在“注释”语法（/ *，//等）： /* multi-line comment */
        # 不常见的字符/运算符： +, *, &, &&, |, ||, <, >, ==, !=, >=, <=, >>, <<, ::, __
        # 帖子中的驼峰文本。
        # 嵌套的括号，大括号。
        num_of_end_char = 0

        af = text.splitlines()
        line_num = len(af)
        for each_line in af:
            if each_line.strip().find(";") > 0:
                num_of_end_char += 1.0
        depth_1 = self.maxDepth(text, "{", "}")
        depth_2 = self.maxDepth(text, "(", ")")
        depth_3 = self.maxDepth(text, "<", ">")
        counter_1 = max(
            text.count("{"), text.count("["), text.count("&&"),
            text.count("||"), text.count(">>"), text.count("<<"),
            text.count("::"), text.count("!="), text.count("=="), text.count(".*"),
        )
        counter_3 = len(self.brackets_pattern.findall(text))
        counter_4 = self.comment_nums(text)
        counter_5 = max(len(self.arrow_pattern.findall(text)), len(self.point_pattern.findall(text)))

        camel_list = self.all_camel_case(text)
        if text.find("implements ") > 0 or text.find("extends ") > 0 and depth_3 == 1:
            return True
        if (depth_1 >= 1 or depth_2 >= 2) and num_of_end_char / line_num > 0.2:
            return True
        if num_of_end_char / line_num > 0.3:
            return True
        if counter_3 >= 2 and text.find("════") < 0:
            return True
        if counter_4 >= 1:
            return True
        if (depth_1 >= 1 or depth_2 >= 2) and len(camel_list) >= 2 and counter_3 >= 1 and text.find("════") < 0:
            return True
        if counter_5 >= 2 and num_of_end_char / line_num > 0.1:
            return True
        if counter_1 >= 5 and num_of_end_char / line_num > 0.1:
            return True
        if counter_1 / line_num >= 0.1 and len(camel_list) >= 2:
            return True
        return False

    def maxDepth(sefl, S, p, q):
        current_max = 0
        max_d = 0
        n = len(S)
        # Traverse the input string
        for i in range(n):
            if S[i] == p:
                current_max += 1
                if current_max > max_d:
                    max_d = current_max
            elif S[i] == q:
                if current_max > 0:
                    current_max -= 1
                else:
                    return -1
        return max_d

    def comment_nums(self, text):
        m = self.comment_pattern.findall(text)
        return len(m)



from unittest import TestCase

from sekg.wiki.search_domain_wiki.wikidata_searcher import AsyncWikiSearcher


class TestWikiSearcher(TestCase):
    def test_update_cache_usage_example(self):
        # 以下代码和提前缓存有关，基本上就是mo
        searcher = AsyncWikiSearcher(proxy_server="http://127.0.0.1:1080")
        title_path = "title.bin"
        item_path = "item.bin"
        # 加载缓存
        searcher.init_from_cache(title_save_path=title_path, item_save_path=item_path)
        # 根据制定标题提前进行查询，这样就会缓存下来
        titles=["java","json","image file"]
        searcher.search_title(titles)

        # 多次执行缓存操作，因为网络可能会波动，导致查询失败，但是多次执行基本上就全部能缓存到了
        for i in range(1,5):
            # 将所有与已经搜索过的标题相关的wikidata详细进行缓存，以便查询
            searcher.fetch_item_for_cache_title_search()

        # 将目前缓存的wikidata节点的邻居也进行缓存，以便需要
        searcher.fetch_item_neighbor()

        # 缓存指定ID的wikidata节点
        searcher.fetch_item(["Q274","Q3434"])

        # 保存title搜索和WikidataItem缓存结果，以便后面离线使用
        searcher.save(title_save_path=title_path,item_save_path=item_path)


    def test_wiki_searcher(self):
        test_set = set()
        test_set.add("Peer alarm")
        test_set.add("Keyboard")
        test_set.add("File")
        test_set.add("Java")
        test_set.add("keyboard")
        test_set.add('Shoutbox')
        searcher = AsyncWikiSearcher(proxy_server="http://127.0.0.1:1080")
        ids = ['Q1670525']
        titles = test_set
        searcher.search_title(titles)
        print(searcher.get_title_cache())
        title_path = "title.bin"
        searcher.save(title_save_path=title_path)
        searcher.update(title_path, type=searcher.TYPE_TITLE)

        searcher.fetch_item(ids)
        print(searcher.get_item_cache())
        item_path = "item.bin"
        searcher.save(item_save_path=item_path)
        searcher.update(item_path, type=searcher.TYPE_ITEM)
        searcher.fetch_item_for_cache_title_search()
        searcher.fetch_item_neighbor(["Q21028"])
        searcher.fetch_item_neighbor()
        print(searcher.get_item_cache())

        searcher.fetch_wikipedia_context(titles)
        print(searcher.get_wikipedia_cache())
        wiki_path = "wikipedia_context.bin"
        searcher.save(wikipedia_content_save_path=wiki_path)
        searcher.update(wiki_path, type=searcher.TYPE_WIKIPEDIA)

        searcher.fetch_wikipedia_context_for_wikidata(ids)
        print(searcher.get_wikidata_wikipedia_context_cache())

        searcher.fetch_wikipedia_context_html(titles)
        print(searcher.get_wikipedia_content_html())

        searcher.fetch_wikidata_item_by_wikipedia_title(["Prototype-based_programming"])
        print("item cache=", searcher.item_cache)
        print("title search cache=", searcher.get_title_cache())

        print("search for Keyboard:", searcher.search_wd_ids_by_keywords("Keyboard"))
        print("search for keyboard:", searcher.search_wd_ids_by_keywords("keyboard"))
        print("search for File:", searcher.search_wd_ids_by_keywords("File"))
        print("search for file:", searcher.search_wd_ids_by_keywords("file"))
        print("search for Java:", searcher.search_wd_ids_by_keywords("Java"))

        print("search for java:", searcher.search_wd_ids_by_keywords("java"))
        print("search for file:", searcher.search_wikidata_items_by_keywords("file"))
        print("search for Keyboard:", searcher.search_wikidata_items_by_keywords("Keyboard"))
        print("search for keyboard:", searcher.search_wikidata_items_by_keywords("keyboard"))
        print("search for Java:", searcher.search_wikidata_items_by_keywords("Java"))



import numpy as np
from gensim import matutils
from numpy import dot, array

"""
The methods are copied from gensim.models.keyedvectors.Doc2VecKeyedVectors in gensim for compute the vector similarity
"""


class VectorUtil:
    @staticmethod
    def similarity(vector_1, vector_2):
        """Compute cosine similarities between one vector and another vector.

            Parameters
            ----------
            vector_1 : numpy.ndarray
                    Vector from which similarities are to be computed, expected shape (dim,).
            vector_2 : numpy.ndarray
                    Vector from which similarities are to be computed, expected shape (dim,).

            Returns
            -------
            numpy.ndarray
            Contains cosine distance between `vector_1` and `vector_2`

        """

        return dot(matutils.unitvec(vector_1), matutils.unitvec(vector_2))

    @staticmethod
    def cosine_similarities(vector_1, vectors_all):
        """Compute cosine similarities between one vector and a set of other vectors.

        Parameters
        ----------
        vector_1 : numpy.ndarray
            Vector from which similarities are to be computed, expected shape (dim,).
        vectors_all : list of numpy.ndarray
            For each row in vectors_all, distance from vector_1 is computed, expected shape (num_vectors, dim).

        Returns
        -------
        numpy.ndarray
            Contains cosine distance between `vector_1` and each row in `vectors_all`, shape (num_vectors,).

        """
        norm = np.linalg.norm(vector_1)
        all_norms = np.linalg.norm(vectors_all, axis=1)
        dot_products = dot(vectors_all, vector_1)
        similarities = dot_products / (norm * all_norms)
        return similarities

    @staticmethod
    def n_similarity(vector_list_1, vector_list_2):
        """Compute cosine similarity between two sets of vectors.

        Parameters
        ----------
        vector_list_1 : list of numpy vector
        vector_list_2: list of numpy vector

        Returns
        -------
        numpy.ndarray
            Similarities between vector_list_1 and vector_list_2.

        """
        if not (len(vector_list_1) and len(vector_list_1)):
            raise ZeroDivisionError('At least one of the passed list is empty.')

        return dot(matutils.unitvec(array(vector_list_1).mean(axis=0)),
                   matutils.unitvec(array(vector_list_2).mean(axis=0)))

    @staticmethod
    def get_weight_mean_vec(vector_list, weight_list):
        """
        get the average word2vec for list of str
        :param weight_list:
        :param vector_list:
        :return: np.array()
        """
        # todo: add a empty zero vectors result.
        # weight_sum = np.sum(weight_list)
        # normal_weight_list = []
        # for w in weight_list:
        #     normal_weight_list.append(w / weight_sum)
        # weight_list=normal_weight_list

        x = np.matrix(vector_list)
        avg_vector = np.average(x, axis=0, weights=weight_list)
        avg_vector = avg_vector.getA()[0]
        return avg_vector

    @staticmethod
    def compute_idf_weight_dict(total_num, number_dict):
        index_2_key_map = {}

        index = 0

        count_list = []
        for key, count in number_dict.items():
            index_2_key_map[index] = key
            count_list.append(count)
            index = index + 1

        a = np.array(count_list)
        ## smooth, in case the divide by zero error
        a = np.log((total_num + 1) / (a + 1))
        result = {}

        for index, w in enumerate(a):
            key = index_2_key_map[index]
            result[key] = w

        return result



# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: attenton
@Email: 18212010081@fudan.edu.cn
@Created: 2019/12/23
------------------------------------------
@Modify: 2019/12/23
------------------------------------------
@Description:
"""
import numpy as np
import pandas as pd
from pathlib import Path
from sklearn.metrics import mean_absolute_error
from sklearn.metrics import cohen_kappa_score


# sklearn中也有cohen_kappa_score方法可以计算cohen kappa，只是输入不同，sklearn的输入是两个向量，而不是矩阵
#

def kappa(testData, k):
    """
    sklearn中也有计算cohern_kappa的，只是输入不同，sklearn输入的两个向量
    :param testData:  需要计算的矩阵，k*k,
    :param k: k表示标注的label个数
    :return:
    """
    dataMat = np.mat(testData)
    P0 = 0.0
    for i in range(k):
        P0 += dataMat[i, i]*1.0
    xsum = np.sum(dataMat, axis=1)
    ysum = np.sum(dataMat, axis=0)
    #xsum是个k行1列的向量，ysum是个1行k列的向量
    sum = np.sum(dataMat)
    # print(sum)
    # print(xsum)
    # print(ysum)
    Pe  = float(ysum*xsum)/sum/sum
    # print(Pe)
    P0 = float(P0/sum*1.0)
    # print(P0)
    cohens_coefficient = float((P0-Pe)/(1-Pe))
    return cohens_coefficient


def fleiss_kappa(testData, N, k, n): #testData表示要计算的数据，（N,k）表示矩阵的形状，说明数据是N行k列的，一共有n个标注人员
    """

    :param testData: 输入的数据矩阵， N*k
    :param N: N条标注的数据
    :param k: k表示标注的label的数量
    :param n: n表示标注人员的数量
    :return:
    """
    dataMat = np.mat(testData, float)
    oneMat = np.ones((k, 1))
    sum = 0.0
    P0 = 0.0
    for i in range(N):
        temp = 0.0
        for j in range(k):
            sum += dataMat[i, j]
            temp += 1.0*dataMat[i, j]**2
        temp -= n
        temp /= (n-1)*n
        P0 += temp
    P0 = 1.0*P0/N
    ysum = np.sum(dataMat, axis=0)
    for i in range(k):
        ysum[0, i] = (ysum[0, i]/sum)**2
    Pe = ysum*oneMat*1.0
    ans = (P0-Pe)/(1-Pe)
    return ans[0, 0]



# -*- coding: utf-8 -*-

class Words2Numbers:
    @staticmethod
    def text2int(textnum, numwords={}):
        if not numwords:
            units = [
                "zero", "one", "two", "three", "four", "five", "six", "seven", "eight",
                "nine", "ten", "eleven", "twelve", "thirteen", "fourteen", "fifteen",
                "sixteen", "seventeen", "eighteen", "nineteen",
            ]
            tens = ["", "", "twenty", "thirty", "forty", "fifty", "sixty", "seventy", "eighty", "ninety"]

            scales = ["hundred", "thousand", "million", "billion", "trillion"]

            numwords["and"] = (1, 0)
            for idx, word in enumerate(units):  numwords[word] = (1, idx)
            for idx, word in enumerate(tens):       numwords[word] = (1, idx * 10)
            for idx, word in enumerate(scales): numwords[word] = (10 ** (idx * 3 or 2), 0)

        ordinal_words = {'first': 1, 'second': 2, 'third': 3, 'fifth': 5, 'eighth': 8, 'ninth': 9, 'twelfth': 12}
        ordinal_endings = [('ieth', 'y'), ('th', '')]

        textnum = textnum.replace('-', ' ')

        current = result = 0
        curstring = ""
        onnumber = False
        for word in textnum.split():
            if word in ordinal_words:
                scale, increment = (1, ordinal_words[word])
                current = current * scale + increment
                if scale > 100:
                    result += current
                    current = 0
                onnumber = True
            else:
                for ending, replacement in ordinal_endings:
                    if word.endswith(ending):
                        word = "%s%s" % (word[:-len(ending)], replacement)

                if word not in numwords:
                    if onnumber:
                        curstring += repr(result + current) + " "
                    curstring += word + " "
                    result = current = 0
                    onnumber = False
                else:
                    scale, increment = numwords[word]

                    current = current * scale + increment
                    if scale > 100:
                        result += current
                        current = 0
                    onnumber = True

        if onnumber:
            curstring += repr(result + current)

        return curstring

if __name__ == '__main__':
    print(Words2Numbers.text2int('80 characters'))


import cProfile
import time
import traceback
from functools import wraps


def conn_try_again(function):
    retries = 5
    # retry time
    count = {"num": retries}

    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as err:
            if count['num'] < 2:
                count['num'] += 1
                return wrapped(*args, **kwargs)
            else:
                raise Exception(err)

    return wrapped


def catch_exception(function):
    def wrapped(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except Exception as err:
            traceback.print_exc()

    return wrapped


def exeTime(function):
    def wrapped(*args, **kwargs):
        t0 = time.time()
        print("@%s, {%s} start" % (time.strftime("%X", time.localtime()), function.__name__))
        back = function(*args, **kwargs)
        print("@%s, {%s} end" % (time.strftime("%X", time.localtime()), function.__name__))
        print("@%.3fs taken for {%s}" % (time.time() - t0, function.__name__))
        return back

    return wrapped


def statsProfile(function):
    """
    statics the detail of method run
    :param function:
    :return:
    """

    def wrapped(*args, **kwargs):
        profile = cProfile.Profile()
        profile.enable()
        back = function(*args, **kwargs)
        profile.disable()
        profile.print_stats(sort="tottime")

        return back

    return wrapped


class FunctionResultCacher:
    func_result_dict = {}
    # todo, there are another tools that support cache result by specified the number of cacher.
    # todo, implement.
    """
    From this page: https://www.cnblogs.com/ydf0509/p/9341365.html
    
    {
        (f1,(1,2,3,4)):(10,1532066199.739),
        (f2,(5,6,7,8)):(26,1532066211.645),
    }
    """

    @classmethod
    def cached_function_result_for_a_time(cls, cache_time):
        """
        函数的结果缓存一段时间装饰器
        :param cache_time 缓存的时间
        :type cache_time : float
        """

        def _cached_function_result_for_a_time(fun):

            @wraps(fun)
            def __cached_function_result_for_a_time(*args, **kwargs):
                if len(cls.func_result_dict) > 1024:
                    cls.func_result_dict.clear()

                key = cls._make_arguments_to_key(args, kwargs)
                if (fun, key) in cls.func_result_dict and time.time() - cls.func_result_dict[(fun, key)][
                    1] < cache_time:
                    return cls.func_result_dict[(fun, key)][0]
                else:
                    result = fun(*args, **kwargs)
                    cls.func_result_dict[(fun, key)] = (result, time.time())
                    # cls.logger.debug('函数 [{}] 此次不使用缓存'.format(fun.__name__))
                    return result

            return __cached_function_result_for_a_time

        return _cached_function_result_for_a_time

    @staticmethod
    def _make_arguments_to_key(args, kwds):
        key = args
        if kwds:
            sorted_items = sorted(kwds.items())
            for item in sorted_items:
                key += item
        return key



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: isky
@Email: 19110240019@fudan.edu.cn
@Created: 2020/02/20
------------------------------------------
@Modify: 2020/02/20
------------------------------------------
@Description:
"""
from pathlib import Path
from unittest import TestCase

from sekg.model.word2vec.tune_word2vec import TunedWord2VecTrainer
from test.data.definition import ROOT_DIR
from test.ir.models.doc_sim_test_data import FakeDocumentCollectionData


class TestTuneWord2vec(TestCase):
    def test_tune(self):
        dc = FakeDocumentCollectionData.fake_pre_dc()
        word2vec_path = Path(ROOT_DIR) / "pretrain" / "pretrainwiki.100.w2v.bin"
        model = TunedWord2VecTrainer.tune_from_pre_doc_collection(dc, pretrain_w2v_path=str(word2vec_path))



from sekg.graph.exporter.graph_data import GraphData
from sekg.graph.exporter.scheme_graph_data import SchemaGraphData
from sekg.ir.doc.wrapper import MultiFieldDocumentCollection
import matplotlib.pyplot as plt
from networkx import nx


class SchemaGraphUtil:
    """
    A tool to create schemaGraph based on GraphData/DocumentCollection
    """
    Entity_Type_Name_Prefix = "entity_type_"
    Relation_Has_Label = "has label"
    Relation_Has_Property = "has property"
    Relation_Has_Field = "has field"
    Relation_Start_Relation = "start"
    Relation_End_Relation = "end"
    label_2_color_map = {"entity_type": 1, "label": 0.8231, "document_field": 0.3714285714285714, "relation": 0.7,
                         "property": 0.66}

    def __init__(self, graph_data: GraphData = None, doc_collection: MultiFieldDocumentCollection = None, save_path=""):
        self.graph_data = graph_data
        self.doc_collection = doc_collection
        self.schema_graph = SchemaGraphData()
        self.schema_save_path = save_path
        self.entity_type_entity_id_index = 0
        self.entity_type_entity_key_id_map = dict()
        self.node_id_2_entity_type = dict()

    def create_related_index(self):
        self.schema_graph.create_index_on_property("unique_name")

    def create_relation_entity_nodes(self):
        relation_set = self.graph_data.get_all_relation_types()
        for relation_name in relation_set:
            self.schema_graph.add_relation_entity_node(relation_name)
        return True

    def create_doc_filed_entity_nodes(self):
        if self.doc_collection is None:
            print("doc_collection is None!")
            return
        for field_name in self.doc_collection.get_field_set():
            self.schema_graph.add_document_field_entity_node(field_name)
        return True

    def create_all_labels(self):
        label_set = self.graph_data.get_all_labels()
        for label in label_set:
            self.schema_graph.add_label_entity_node(label)
        return True

    def add_all_property_nodes(self):
        for node_id, node_json in self.graph_data.graph.nodes(data=True):
            if GraphData.DEFAULT_KEY_NODE_PROPERTIES in node_json:
                properties = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]
                for p in properties:
                    self.schema_graph.add_property_entity_node(p)
        return True

    def get_key(self, label_set):
        key = ""
        for label in label_set:
            key += (label + "&")
        return key

    def create_all_entity_types(self):
        """
        unique label set represent one entity type
        """
        for node_id, node_json in self.graph_data.graph.nodes(data=True):
            if GraphData.DEFAULT_KEY_NODE_LABELS in node_json:
                label_set = node_json[GraphData.DEFAULT_KEY_NODE_LABELS]
                key = self.get_key(label_set)
                if key in self.entity_type_entity_key_id_map:
                    self.node_id_2_entity_type[node_id] = self.entity_type_entity_key_id_map[key]
                    continue
                else:
                    self.entity_type_entity_key_id_map[key] = SchemaGraphUtil.Entity_Type_Name_Prefix + str(
                        self.entity_type_entity_id_index)
                    self.node_id_2_entity_type[node_id] = self.entity_type_entity_key_id_map[key]
                    self.entity_type_entity_id_index += 1
                    self.schema_graph.add_entity_type_entity_node(self.entity_type_entity_key_id_map[key])
        return True

    def link_entity_type_2_relations(self):
        all_relation_pairs = self.graph_data.get_relations()
        for start_node_id, relation_name, end_node_id in all_relation_pairs:
            relation_entity_node = self.schema_graph.find_relation_entity_node_by_name_en(relation_name)
            if relation_entity_node is None:
                print("relation_entity_node is None")
                continue
            start_entity_type = self.node_id_2_entity_type[start_node_id]
            start_entity_node = self.schema_graph.find_entity_type_entity_node_by_name_en(start_entity_type)
            if start_entity_type is None:
                print("start_entity_type is None")

            self.schema_graph.add_relation(start_entity_node["id"], SchemaGraphUtil.Relation_Start_Relation,
                                           relation_entity_node["id"])
            end_entity_type = self.node_id_2_entity_type[end_node_id]
            if end_entity_type is None:
                print("end_entity_type is None")
            end_entity_node = self.schema_graph.find_entity_type_entity_node_by_name_en(end_entity_type)
            self.schema_graph.add_relation(end_entity_node["id"], SchemaGraphUtil.Relation_End_Relation,
                                           relation_entity_node["id"])

    def link_entity_type_2_property_label(self):
        for node_id, node_json in self.graph_data.graph.nodes(data=True):
            if GraphData.DEFAULT_KEY_NODE_LABELS in node_json and GraphData.DEFAULT_KEY_NODE_PROPERTIES in node_json:
                label_set = node_json[GraphData.DEFAULT_KEY_NODE_LABELS]
                property_set = node_json[GraphData.DEFAULT_KEY_NODE_PROPERTIES]
                key = self.get_key(label_set)
                if key not in self.entity_type_entity_key_id_map:
                    print("must create_all_entity_types first")
                    continue
                entity_type_json = self.schema_graph.find_entity_type_entity_node_by_name_en(
                    self.entity_type_entity_key_id_map[key])
                if entity_type_json is None:
                    print("not find entity type")
                    continue
                for label in label_set:
                    label_node_json = self.schema_graph.find_label_entity_node_by_name_en(label)
                    if label_node_json is None:
                        print("not find label node")
                        continue
                    self.schema_graph.add_relation(entity_type_json["id"], SchemaGraphUtil.Relation_Has_Label,
                                                   label_node_json["id"])

                for property_name in property_set:
                    property_node_json = self.schema_graph.find_property_entity_node_by_name_en(property_name)
                    if property_node_json is None:
                        print("not find property node")
                        continue
                    self.schema_graph.add_relation(entity_type_json["id"],
                                                   SchemaGraphUtil.Relation_Has_Property,
                                                   property_node_json["id"])

    def link_entity_type_2_doc_filed(self):
        document_list = self.doc_collection.get_document_list()
        for i, document in enumerate(document_list):
            node_id = document.id
            field_set = document.get_field_set()
            entity_type_name = self.node_id_2_entity_type[node_id]
            entity_type_json = self.schema_graph.find_entity_type_entity_node_by_name_en(entity_type_name)
            if entity_type_json is None:
                print("not find entity type")
                continue

            for field in field_set:
                document_field_entity = self.schema_graph.find_document_field_entity_node_by_name_en(field)
                if document_field_entity is None:
                    print("document_field_entity is None")
                    continue
                self.schema_graph.add_relation(entity_type_json["id"], SchemaGraphUtil.Relation_Has_Field,
                                               document_field_entity["id"])

    def create_from_source(self):
        """
        try to create all node info from Graph and Doc
        :return:
        """
        self.add_all_property_nodes()
        self.create_relation_entity_nodes()
        self.create_doc_filed_entity_nodes()
        self.create_all_labels()
        self.create_all_entity_types()
        self.link_all_relation()

    def create_from_graph(self):
        """
        try to create all node info from Graph
        :return:
        """
        self.add_all_property_nodes()
        self.create_relation_entity_nodes()
        self.create_all_labels()
        self.create_all_entity_types()

    def create_from_doc(self):
        self.create_doc_filed_entity_nodes()

    def link_all_relation(self):
        self.link_graph_relation()
        self.link_doc_relation()

    def link_graph_relation(self):
        self.link_entity_type_2_property_label()
        self.link_entity_type_2_relations()

    def link_doc_relation(self):
        self.link_entity_type_2_doc_filed()

    def draw_schema_graph(self):
        G = self.schema_graph.graph
        labels = dict()
        val_map = dict()
        for n, d in G.nodes(data=True):
            labels[n] = d[SchemaGraphData.DEFAULT_KEY_NODE_PROPERTIES][
                SchemaGraphData.DEFAULT_KEY_PROPERTY_QUALIFIED_NAME]
            val_map[n] = SchemaGraphUtil.label_2_color_map[list(d[SchemaGraphData.DEFAULT_KEY_NODE_LABELS])[0]]
        values = [val_map.get(node, 0.25) for node in G.nodes()]
        pos = nx.spring_layout(G)
        edge_labels = {}

        all_relations = self.schema_graph.get_relations()
        for s, r, e in all_relations:
            edge_labels[(s, e)] = r

        nx.draw(G, pos=pos, labels=labels, alpha=0.9, node_size=1000,
                node_color=values,
                cmap=plt.get_cmap('viridis'), font_size=10,
                )
        nx.draw_networkx_edge_labels(G, pos, edge_labels=edge_labels, font_color='red', font_size=8)
        plt.show()

    def save(self):
        self.schema_graph.save(self.schema_save_path)



# -*- coding: utf-8 -*-
"""
-----------------------------------------
@Author: attenton
@Email: 18212010081@fudan.edu.cn
@Created: 2019/12/13
------------------------------------------
@Modify: 2019/12/24
------------------------------------------
@Description:
"""

import spacy
from spacy.lang.char_classes import ALPHA, HYPHENS, ALPHA_LOWER, ALPHA_UPPER, CONCAT_QUOTES

from sekg.text.spacy_pipeline.sentenceHandler import SentenceHandler


class CustomizeSpacy:
    """
    apcy中的tokenizer主要原理分为五个部分，例外rules（也可以被称为tokenizer_exceptions）， 前缀prefix，后缀suffix，中缀infix，以及例外的正则匹配token_match
    修改以上任意一个都可以修改spacy中的tokenizer，实现自定义的tokenizer
    可以使用nlp.tokenizer.explain()查看每一个token的成分，是suffix，inffix，prefix还是token_match
    """

    @staticmethod
    def customize_tokenizer_split_single_lowcase_letter_and_period(nlp):
        """
        自定义的tokenzier，修改spacy中的exception，
        spacy中的tokenizer exception添加了{a.|b.|c.}等单个字母与句号连接的例外，这个可能导致分句出现错误，比如 1 < n < r. 出现在句尾。
        :param nlp:
        :return:
        """
        # prefix_re = spacy.util.compile_prefix_regex(nlp.Defaults.prefixes)
        # suffix_re = spacy.util.compile_suffix_regex(nlp.Defaults.suffixes)
        # infix_re = spacy.util.compile_infix_regex(nlp.Defaults.infixes)
        #
        # # remove all exceptions where a single letter is followed by a period (e.g. 'h.')
        exceptions = {k: v for k, v in dict(nlp.Defaults.tokenizer_exceptions).items() if
                      not (len(k) == 2 and k[1] == '.')}
        # 添加特殊的tokenizer exception
        for orth in [
            "C#",
            "c#"
        ]:
            exceptions[orth] = [{"ORTH": orth}]
        # new_tokenizer = Tokenizer(nlp.vocab, exceptions,
        #                           prefix_search=prefix_re.search,
        #                           suffix_search=suffix_re.search,
        #                           infix_finditer=infix_re.finditer,
        #                           token_match=nlp.tokenizer.token_match)
        #
        # nlp.tokenizer = new_tokenizer
        nlp.tokenizer.rules = exceptions

    @staticmethod
    def customize_tokenizer_merge_hyphen(nlp):
        """
        将连字符相连的合并为一个token， 如原来的fail-fast在tokenize之后会分成fail - fast三个，修改之后会全部合并为fail-fast这一个token
        实现方法： 删除infix中分离连字符的正则表达式
        :param nlp:
        :return:
        """
        _infixes = list(nlp.Defaults.infixes)
        _infixes.remove(r"(?<=[{a}])(?:{h})(?=[{a}])".format(a=ALPHA, h=HYPHENS), )
        infix_regex = spacy.util.compile_infix_regex(_infixes)
        nlp.tokenizer.infix_finditer = infix_regex.finditer

    @staticmethod
    def customize_tokenizer_merge_dot_upper_letter(nlp):
        """
        将.{upper_letter}相连的合并为一个token， 如原来的java.String.format在tokenize之前会被拆分为`java.` `String.format`
        修改tokenizer之后会合并为`java.String.format`一个
        实现方法： 删除infix中.{upper_letter}的正则表达式
        :param nlp:
        :return:
        """
        _infixes = list(nlp.Defaults.infixes)
        _infixes.remove(r"(?<=[{al}{q}])\.(?=[{au}{q}])".format(
            al=ALPHA_LOWER, au=ALPHA_UPPER, q=CONCAT_QUOTES
        ), )
        infix_regex = spacy.util.compile_infix_regex(_infixes)
        nlp.tokenizer.infix_finditer = infix_regex.finditer

    @staticmethod
    def customize_tokenizer_api_name_recognition(nlp):
        """
        将句子中的api名字识别成一个token, 在生成tokenizer的函数后加上token.match
        将中括号的内容合并，之前会将]作为单独的token
        TODO: 目前只是将中括号的合并了，由于token_match是在suffix、prefix之后执行，所以复杂的正则会被suffix、prefixoverride，所以目前还是用doc.retoeknizer()方法实现api_name的识别
        :param nlp:
        :return:
        """
        # re_token_match = _get_regex_pattern(nlp.Defaults.token_match)
        _suffixes = list(nlp.Defaults.suffixes)
        _suffixes.remove("\]")
        suffixes_regex = spacy.util.compile_suffix_regex(_suffixes)
        nlp.tokenizer.suffix_search = suffixes_regex.search
        # api_name_match = re.compile("(\w+(\[[\w+-]+\])?\.)+\w+\(.*?\)", re.UNICODE).match
        # token_match = nlp.Defaults.token_match
        # print(token_match)
        # token_match.append(api_name_match)
        # nlp.tokenizer.token_match = api_name_match

    @staticmethod
    def customize_sentencizer_merge_colon(doc):
        """
        之前碰到冒号会分句，处理之后不再会分句
        :param doc:
        :return:
        """
        for token in doc[:-1]:
            if token.text == ":":
                doc[token.i+1].is_sent_start = False
        return doc

    @staticmethod
    def pipeline_merge_bracket(doc):
        """
        Java int to String - Integer.toString(i) vs new Integer(i).toString
        --> ['Java', 'int', 'to', 'String', '-', 'Integer.toString(i)', 'vs', 'new', 'Integer(i).toString']
        将括号合并： 如果'('的左边是空格，那么就括号内的内容自身合并， 如果左边不是空格，那么就和左边的内容一起合并，比如Integer.toString(i)
        :param doc:
        :return:
        """
        method_brace_pair_list = []
        method_left_brace_cache_index = -1
        left_angle_bracket = -1
        angle_brackets_list = []
        count = 0
        for i, token in enumerate(doc):
            # todo" complete this for sentence containing method name
            # 添加了对method name的处理
            if "<" == token.text:
                if i > 0 and doc[i-1].text != ">":
                    left_angle_bracket = i-1
                else:
                    left_angle_bracket = i
            elif "<" in token.text:
                left_angle_bracket = i

            if ">" == token.text and left_angle_bracket != -1:
                angle_brackets_list.append((left_angle_bracket, i))
                left_angle_bracket = -1
        # print(method_brace_pair_list)
        with doc.retokenize() as retokenizer:
            for star_index, end_index in angle_brackets_list:
                attrs = {"LEMMA": " ".join([token.text for token in doc[(star_index):(end_index+1)]]),
                         "POS": "NOUN",
                         "TAG": "NN",
                         }
                retokenizer.merge(doc[star_index:end_index+1], attrs=attrs)
        for i, token in enumerate(doc):
            # todo" complete this for sentence containing method name
            # 添加了对method name的处理
            if "(" == token.text:
                if count == 0:
                    # 判断'('在原句中和前一个token有没有空格，如果没有空格就与前一个token一起合并
                    if i > 0 and doc[i - 1].text != ")" and doc[i - 1].text != "." and doc.text.find(doc[i - 1].text+'(') != -1:
                        method_left_brace_cache_index = i-1
                    else:
                        method_left_brace_cache_index = i
                count += 1
            # 有些时候'('和其他一些单词一起组成一个token
            elif "(" in token.text:
                if count == 0:
                    method_left_brace_cache_index = i
                count += 1
            if ")" == token.text and method_left_brace_cache_index != -1:
                count -= 1
                if count == 0:
                    method_brace_pair_list.append((method_left_brace_cache_index, i))
                    method_left_brace_cache_index = -1
        # print(method_brace_pair_list)
        with doc.retokenize() as retokenizer:
            for star_index, end_index in method_brace_pair_list:
                span = doc[star_index:(end_index + 1)]
                if not SentenceHandler.is_complete_sentence(span):
                    attrs = {"LEMMA": " ".join([token.text for token in doc[(star_index):(end_index + 1)]]),
                             "POS": "NOUN",
                             "TAG": "NN",
                             }
                    retokenizer.merge(doc[star_index:end_index + 1], attrs=attrs)

        return doc


if __name__ == '__main__':
    nlp = spacy.load('en_core_web_sm')
    CustomizeSpacy.customize_tokenizer_split_single_lowcase_letter_and_period(nlp)
    CustomizeSpacy.customize_tokenizer_merge_hyphen(nlp)
    CustomizeSpacy.customize_tokenizer_merge_dot_upper_letter(nlp)
    CustomizeSpacy.customize_tokenizer_api_name_recognition(nlp)
    sentencizer = nlp.create_pipe("sentencizer")
    nlp.add_pipe(sentencizer, before="tagger")
    nlp.add_pipe(CustomizeSpacy.customize_sentencizer_merge_colon, before="tagger")
    # nlp.add_pipe(SentenceHandler.hyphen_handler, name='sh', after='tagger')
    nlp.add_pipe(CustomizeSpacy.pipeline_merge_bracket, name='pipeline_merge_bracket', after='tagger')
    # sentence = 'java.lang.StringBuffer create StringBuffer'
    # sentence = 'Suppose that a byte sequence of length n is written, where < n <= r. Up to the first srcs[offset].remaining() bytes of this sequence are written from buffer srcs[offset], up to the next srcs[offset+1].remaining() bytes are written from buffer srcs[offset+1], and so forth, until the entire byte sequence is written.'
    sentences = ['How to use java.String.format in Scala?',
                 'How to use java.string.format in Scala?',
                 'How to use (use) dfads.',
                 'However, I get a NoSuchElementException (since there is no invocation of hasNext) -CODE- .',
                 'Why is 2 * (i * i) faster than 2 * i * i in Java?',
                 'Why is 2 * i Java?',
                 'In C# it is DateTime.Now.',
                 "I was really surprised but I haven't been able to find anything that shows how to get the MD5 checksum of a file.",
                 "you get the className + @ + the hex of the hashCode of the array, as defined by Object.toString(): -CODE- .",
                 "I get this: -CODE- .",
                 "I'm trying to remove some elements from an ArrayList while iterating it like this: -CODE- .",
                 "There are some obvious ways of doing this (i.e. with a loop) but I would like something a bit neater, something like: -CODE- .",
                 "Equivalent of C#'s DateTime.Now in Java?",
                 "Java int to String - Integer.toString(i) vs new Integer(i).toString()",
                 "What is the fundamental difference between the Set<E> and List<E> interfaces?",
                 "What issues / pitfalls must be considered when overriding equals and hashCode?",
                 "In the following program you can see that each value slightly less than .5 is rounded down, except for 0.5.",
                 "I can accomplish this easily using Java 8 and Guava but I would like to know how to do this without Guava.",
                 "However, when I deploy in WebSphere I get a java.lang.NoClassDefFoundError: org.slf4j.impl.StaticLoggerBinder.",
                 "Method one: static initialiser Method two: instance initialiser (anonymous subclass) or some other method?",
                 "What I would like to know is whether I should use System.currentTimeMillis() or System.nanoTime() when updating my object's positions in my game?",
                 "I get this: -CODE- and I would like to find the enum value of a string, for example \"A\" which would be Blah.A.",
                 "Are System.out.printf and System.out.format totally the same or perhaps they differ in somehow? Are printf and format totally the same or perhaps they differ in somehow?",
                 "Are printf and format totally the same or perhaps they differ in somehow?",
                 "Retrieves a value set in the state of this engine. The value might be one which was set using setValue or some other value in the state of the ScriptEngine, depending on the implementation. Must have the same effect as getBindings(ScriptContext.ENGINE_SCOPE).get"
                 ]
    # sentence = 'How do I get the current stacktrace in Java, like how in .NET you can do Environment.StackTrace'
    # sentence = "The elements in the array returned by the method executeBatch may be one of the following: A number greater than or equal to zero -- indicates that the command was processed successfully and is an update count giving the number of rows in the database that were affected by the command's execution"
    # sentence = "The maximum number of buffers-dke-dfs to be accessed 1 < n < r. So API must be non-negative and no larger than srcs.length - offset Returns long The number of bytes written, possibly zero"
    for sentence in sentences:
        doc = nlp(sentence)
        # debug tokenizer
        tok_exp = nlp.tokenizer.explain(sentence)
        for t in tok_exp:
            print(t[1], "\t", t[0])
        for i in doc.sents:
            print(i)
            print([token.text for token in i])
            print([token.dep_ for token in i])
            print([token.tag_ for token in i])
            print('\n')
        print('-------------------------')
        print('\n')



from sekg.graph.exporter.graph_data import GraphData


class BaseEntity:
    def __init__(self, name_en, name_zh, declaration, example):
        self.name_en = name_en
        self.name_zh = name_zh
        self.declaration = declaration
        self.example = example

    def get_all_properties(self):
        return {
            "name_en": self.name_en,
            "name_zh": self.name_zh,
            "declaration": self.declaration,
            "example": self.example
        }


class PropertyEntity(BaseEntity):
    name_prefix = "Property:"

    def __init__(self, name_en, name_zh="", declaration="", example=""):
        BaseEntity.__init__(self, name_en=name_en, name_zh=name_zh,
                            declaration=declaration, example=example)
        self.label = "property"

    def get_all_properties(self):
        res = BaseEntity.get_all_properties(self)
        res["unique_name"] = PropertyEntity.name_prefix + self.name_en
        return res


class RelationEntity(BaseEntity):
    name_prefix = "Relation:"

    def __init__(self, name_en, name_zh="", declaration="", example=""):
        BaseEntity.__init__(self, name_en=name_en, name_zh=name_zh,
                            declaration=declaration, example=example)
        self.label = "relation"

    def get_all_properties(self):
        res = BaseEntity.get_all_properties(self)
        res["unique_name"] = RelationEntity.name_prefix + self.name_en
        return res


class DocumentFieldEntity(BaseEntity):
    name_prefix = "DocumentField:"

    def __init__(self, name_en, name_zh="", declaration="", example=""):
        BaseEntity.__init__(self, name_en=name_en, name_zh=name_zh,
                            declaration=declaration, example=example)
        self.label = "document_field"

    def get_all_properties(self):
        res = BaseEntity.get_all_properties(self)
        res["unique_name"] = DocumentFieldEntity.name_prefix + self.name_en
        return res


class LabelEntity(BaseEntity):
    name_prefix = "Label:"

    def __init__(self, name_en, name_zh="", declaration="", example=""):
        BaseEntity.__init__(self, name_en=name_en, name_zh=name_zh,
                            declaration=declaration, example=example)
        self.label = "label"

    def get_all_properties(self):
        res = BaseEntity.get_all_properties(self)
        res["unique_name"] = LabelEntity.name_prefix + self.name_en
        return res


class EntityTypeEntity(BaseEntity):
    name_prefix = "EntityType:"

    def __init__(self, name_en, name_zh="", declaration="", example=""):
        BaseEntity.__init__(self, name_en=name_en, name_zh=name_zh,
                            declaration=declaration, example=example)
        self.label = "entity_type"

    def get_all_properties(self):
        res = BaseEntity.get_all_properties(self)
        res["unique_name"] = EntityTypeEntity.name_prefix + self.name_en
        return res


class SchemaGraphData(GraphData):
    """
    GraphData metadata class, store graph data property entity,relation entity,
    documentation filed entity, tag entity, entity type entity, and so on...

    support visualization
    """
    DEFAULT_KEY_PROPERTY_QUALIFIED_NAME = "unique_name"

    def __init__(self):
        GraphData.__init__(self)

    def add_property_entity_node(self, name_en, name_zh="", declaration="", example=""):
        property_entity = PropertyEntity(name_en, name_zh, declaration, example)
        return self.add_node({property_entity.label}, property_entity.get_all_properties(),
                             primary_property_name="unique_name")

    def find_property_entity_node_by_name_en(self, name_en):
        query_name = PropertyEntity.name_prefix + name_en
        return self.find_one_node_by_property("unique_name", query_name)

    def remove_property_entity_node(self, name_en):
        node_json = self.find_property_entity_node_by_name_en(name_en)
        if node_json is None:
            return None
        return self.remove_node(node_json["id"])

    def add_relation_entity_node(self, name_en, name_zh="", declaration="", example=""):
        entity = RelationEntity(name_en, name_zh, declaration, example)
        return self.add_node({entity.label}, entity.get_all_properties(),
                             primary_property_name="unique_name")

    def find_relation_entity_node_by_name_en(self, name_en):
        query_name = RelationEntity.name_prefix + name_en
        return self.find_one_node_by_property("unique_name", query_name)

    def remove_relation_entity_node(self, name_en):
        node_json = self.find_relation_entity_node_by_name_en(name_en)
        if node_json is None:
            return None
        return self.remove_node(node_json["id"])

    def add_document_field_entity_node(self, name_en, name_zh="", declaration="", example=""):
        entity = DocumentFieldEntity(name_en, name_zh, declaration, example)
        return self.add_node({entity.label}, entity.get_all_properties(),
                             primary_property_name="unique_name")

    def find_document_field_entity_node_by_name_en(self, name_en):
        query_name = DocumentFieldEntity.name_prefix + name_en
        return self.find_one_node_by_property("unique_name", query_name)

    def remove_document_field_entity_node(self, name_en):
        node_json = self.find_document_field_entity_node_by_name_en(name_en)
        if node_json is None:
            return None
        return self.remove_node(node_json["id"])

    def add_label_entity_node(self, name_en, name_zh="", declaration="", example=""):
        entity = LabelEntity(name_en, name_zh, declaration, example)
        return self.add_node({entity.label}, entity.get_all_properties(),
                             primary_property_name="unique_name")

    def find_label_entity_node_by_name_en(self, name_en):
        query_name = LabelEntity.name_prefix + name_en
        return self.find_one_node_by_property("unique_name", query_name)

    def remove_label_filed_entity_node(self, name_en):
        node_json = self.find_label_entity_node_by_name_en(name_en)
        if node_json is None:
            return None
        return self.remove_node(node_json["id"])

    def add_entity_type_entity_node(self, name_en, name_zh="", declaration="", example=""):
        entity = EntityTypeEntity(name_en, name_zh, declaration, example)
        return self.add_node({entity.label}, entity.get_all_properties(),
                             primary_property_name="unique_name")

    def find_entity_type_entity_node_by_name_en(self, name_en):
        query_name = EntityTypeEntity.name_prefix + name_en
        return self.find_one_node_by_property("unique_name", query_name)

    def remove_entity_type_filed_entity_node(self, name_en):
        node_json = self.find_entity_type_entity_node_by_name_en(name_en)
        if node_json is None:
            return None
        return self.remove_node(node_json["id"])
